/********************************************************************************
 * (C) Copyright 2000-2005, by Shawn Qualia.
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this library; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 ********************************************************************************/

package org.uguess.birt.report.engine.emitter.xls;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFClientAnchor;
import org.apache.poi.hssf.usermodel.HSSFComment;
import org.apache.poi.hssf.usermodel.HSSFHyperlink;
import org.apache.poi.hssf.usermodel.HSSFPatriarch;
import org.apache.poi.hssf.usermodel.HSSFPicture;
import org.apache.poi.hssf.usermodel.HSSFRichTextString;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.Region;
import org.eclipse.birt.report.engine.api.IHTMLActionHandler;
import org.eclipse.birt.report.engine.api.IRenderOption;
import org.eclipse.birt.report.engine.api.IReportRunnable;
import org.eclipse.birt.report.engine.api.RenderOption;
import org.eclipse.birt.report.engine.api.RenderOptionBase;
import org.eclipse.birt.report.engine.api.impl.Action;
import org.eclipse.birt.report.engine.content.IContent;
import org.eclipse.birt.report.engine.content.IHyperlinkAction;
import org.eclipse.birt.report.engine.content.IImageContent;
import org.eclipse.birt.report.engine.content.IReportContent;
import org.eclipse.birt.report.engine.emitter.IEmitterServices;
import org.eclipse.birt.report.engine.ir.ReportElementDesign;
import org.eclipse.birt.report.engine.layout.area.IArea;
import org.eclipse.birt.report.engine.layout.area.IAreaVisitor;
import org.eclipse.birt.report.engine.layout.area.IContainerArea;
import org.eclipse.birt.report.engine.layout.area.IImageArea;
import org.eclipse.birt.report.engine.layout.area.ITemplateArea;
import org.eclipse.birt.report.engine.layout.area.ITextArea;
import org.eclipse.birt.report.engine.layout.area.impl.PageArea;
import org.eclipse.birt.report.model.api.ReportElementHandle;
import org.uguess.birt.report.engine.layout.wrapper.Frame;
import org.uguess.birt.report.engine.layout.wrapper.impl.AggregateFrame;
import org.uguess.birt.report.engine.layout.wrapper.impl.AreaFrame;
import org.uguess.birt.report.engine.spreadsheet.model.Cell;
import org.uguess.birt.report.engine.spreadsheet.model.MergeBlock;
import org.uguess.birt.report.engine.spreadsheet.model.Sheet;
import org.uguess.birt.report.engine.spreadsheet.wrapper.Transformer;
import org.uguess.birt.report.engine.util.EngineUtil;
import org.uguess.birt.report.engine.util.ImageUtil;
import org.uguess.birt.report.engine.util.OptionParser;
import org.uguess.birt.report.engine.util.PageRangeChecker;

/**
 * XlsRenderer
 */
public class XlsRenderer implements IAreaVisitor
{

	public static final String DEFAULT_FILE_NAME = "report.xls"; //$NON-NLS-1$

	protected static final String XLS_IDENTIFIER = "xls"; //$NON-NLS-1$

	protected static final int COMMENTS_WIDTH_IN_COLUMN = 3;

	protected static final int COMMENTS_HEIGHT_IN_ROW = 5;

	private static final boolean DEBUG;

	static
	{
		DEBUG = System.getProperty( "debug" ) != null; //$NON-NLS-1$
	}

	protected boolean removeEmptyRow = false;

	protected boolean showGridLines = false;

	protected boolean exportBodyOnly = false;

	protected boolean suppressUnknownImage = false;

	protected boolean exportSingleSheet = false;

	protected boolean enableComments = true;

	protected int fixedColumnWidth = -1;

	private int dpi;

	private double baseCharWidth;

	private AggregateFrame singleContainer;

	private PageRangeChecker rangeChecker;

	private int currentPageIndex;

	private int commentsAnchorColumnIndex, commentsAnchorRowIndex;

	private OutputStream output = null;

	private HSSFWorkbook workbook;

	private XlsStyleProcessor processor;

	protected Stack<Frame> frameStack;

	protected IEmitterServices services;

	private long timeCounter;

	public void initialize( IEmitterServices services )
	{
		// init options.
		rangeChecker = new PageRangeChecker( null );
		removeEmptyRow = false;
		showGridLines = false;
		fixedColumnWidth = -1;
		exportBodyOnly = false;
		exportSingleSheet = false;
		dpi = 96;
		suppressUnknownImage = false;
		enableComments = true;

		this.services = services;

		// parse emitter options.
		Object option = services.getEmitterConfig( ).get( XLS_IDENTIFIER );
		if ( option instanceof Map )
		{
			parseRendererOptions( (Map) option );
		}
		else if ( option instanceof IRenderOption )
		{
			parseRendererOptions( ( (IRenderOption) option ).getOptions( ) );
		}

		// parse rendering options, this will overwrite the option by emitter
		// per rendering task.
		parseRendererOptions( services.getRenderOption( ).getOutputSetting( ) );

		Object fd = services.getOption( RenderOptionBase.OUTPUT_FILE_NAME );
		File file = null;
		try
		{
			if ( fd != null )
			{
				file = new File( fd.toString( ) );
				output = new FileOutputStream( file );
			}
		}
		catch ( FileNotFoundException fnfe )
		{
			fnfe.printStackTrace( );
		}

		if ( output == null )
		{
			Object value = services.getOption( RenderOptionBase.OUTPUT_STREAM );
			if ( value != null && value instanceof OutputStream )
			{
				output = (OutputStream) value;
			}

			else
			{
				try
				{
					file = new File( DEFAULT_FILE_NAME );
					output = new FileOutputStream( file );
				}
				catch ( FileNotFoundException e )
				{
					e.printStackTrace( );
				}
			}
		}
	}

	protected void parseRendererOptions( Map xlsOption )
	{
		if ( xlsOption == null )
		{
			return;
		}

		Object value;

		value = xlsOption.get( "dpi" ); //$NON-NLS-1$
		if ( value != null )
		{
			int pdpi = OptionParser.parseInt( value );

			if ( pdpi != -1 )
			{
				dpi = pdpi;
			}
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_PAGE_RANGE );
		if ( value != null )
		{
			rangeChecker = new PageRangeChecker( OptionParser.parseString( value ) );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_REMOVE_EMPTY_ROW );
		if ( value != null )
		{
			removeEmptyRow = OptionParser.parseBoolean( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_SHOW_GRID_LINES );
		if ( value != null )
		{
			showGridLines = OptionParser.parseBoolean( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_FIXED_COLUMN_WIDTH );
		if ( value != null )
		{
			fixedColumnWidth = OptionParser.parseInt( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_EXPORT_BODY_ONLY );
		if ( value != null )
		{
			exportBodyOnly = OptionParser.parseBoolean( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_EXPORT_SINGLE_SHEET );
		if ( value != null )
		{
			exportSingleSheet = OptionParser.parseBoolean( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_SUPPRESS_UNKNOWN_IMAGE );
		if ( value != null )
		{
			suppressUnknownImage = OptionParser.parseBoolean( value );
		}

		value = xlsOption.get( XlsEmitterConfig.KEY_ENABLE_COMMENTS );
		if ( value != null )
		{
			enableComments = OptionParser.parseBoolean( value );
		}
	}

	public void start( IReportContent rc )
	{
		if ( DEBUG )
		{
			timeCounter = System.currentTimeMillis( );
		}

		this.frameStack = new Stack<Frame>( );

		workbook = new HSSFWorkbook( );

		currentPageIndex = 1;

		baseCharWidth = 7 * 72d / dpi;

		// Font bookFont = null;
		//
		// if ( workbook.getNumberOfFonts( ) > 0 )
		// {
		// HSSFFont font = workbook.getFontAt( (short) 0 );
		//
		// bookFont = new Font( font.getFontName( ),
		// Font.PLAIN,
		// font.getFontHeightInPoints( ) );
		// }
		// else
		// {
		// bookFont = new Font( "Serif", Font.PLAIN, 10 ); //$NON-NLS-1$
		// }
		//
		// Graphics2D g2d = null;
		//
		// try
		// {
		// BufferedImage bi = new BufferedImage( 1,
		// 1,
		// BufferedImage.TYPE_INT_ARGB );
		// g2d = (Graphics2D) bi.getGraphics( );
		// FontMetrics fm = g2d.getFontMetrics( bookFont );
		// double charWidth = fm.charWidth( '0' );
		//
		// baseCharWidth = charWidth;// * dpi / 72d;
		//
		// bi.flush( );
		// }
		// catch ( Exception e )
		// {
		// baseCharWidth = 6;// * dpi / 72d;
		// }
		// finally
		// {
		// if ( g2d != null )
		// {
		// g2d.dispose( );
		// }
		// }

		processor = new XlsStyleProcessor( workbook );
	}

	public void end( IReportContent rc )
	{
		if ( exportSingleSheet && singleContainer != null )
		{
			long span = System.currentTimeMillis( );

			// process single sheet here.
			Sheet modelSheet = Transformer.toSheet( singleContainer,
					1000,
					fixedColumnWidth * 1000 );
			exportSheet( modelSheet );

			if ( DEBUG )
			{
				System.out.println( "------------export total single sheet using " //$NON-NLS-1$
						+ ( System.currentTimeMillis( ) - span )
						+ " ms" ); //$NON-NLS-1$
			}
		}

		try
		{
			workbook.write( output );
		}
		catch ( IOException e )
		{
			e.printStackTrace( );
		}

		frameStack.clear( );

		processor.dispose( );
		processor = null;

		workbook = null;
		frameStack = null;
		singleContainer = null;

		if ( DEBUG )
		{
			System.out.println( "------------total exporting time using: " //$NON-NLS-1$
					+ ( System.currentTimeMillis( ) - timeCounter )
					+ " ms" ); //$NON-NLS-1$
		}
	}

	public void startContainer( IContainerArea container )
	{
		buildFrame( container, true );
	}

	public void endContainer( IContainerArea containerArea )
	{
		Frame currentFrame = frameStack.pop( );

		if ( currentFrame.getData( ) instanceof PageArea )
		{
			exportFrame( currentFrame );
		}
	}

	protected void exportFrame( Frame currentFrame )
	{
		long span = System.currentTimeMillis( );

		if ( rangeChecker.checkRange( currentPageIndex ) )
		{
			if ( exportBodyOnly )
			{
				PageArea pa = (PageArea) currentFrame.getData( );
				IContainerArea body = pa.getBody( );

				// locate body frame first
				Frame bodyFrame = locateChildFrame( currentFrame, body );

				if ( bodyFrame instanceof AreaFrame )
				{
					// adjust offsets
					( (AreaFrame) bodyFrame ).setXOffset( -bodyFrame.getLeft( ) );
					( (AreaFrame) bodyFrame ).setYOffset( -bodyFrame.getTop( ) );
				}

				if ( exportSingleSheet )
				{
					// aggregate only
					if ( singleContainer == null )
					{
						singleContainer = new AggregateFrame( );
						singleContainer.setRight( bodyFrame.getRight( )
								- bodyFrame.getLeft( ) );
						singleContainer.setBottom( bodyFrame.getBottom( )
								- bodyFrame.getTop( ) );
					}
					else
					{
						int yOff = singleContainer.getBottom( );
						singleContainer.setBottom( yOff
								+ bodyFrame.getBottom( )
								- bodyFrame.getTop( ) );

						if ( bodyFrame instanceof AreaFrame )
						{
							( (AreaFrame) bodyFrame ).setYOffset( yOff
									+ ( (AreaFrame) bodyFrame ).getYOffset( ) );
						}
					}

					singleContainer.addChild( bodyFrame );
					if ( bodyFrame instanceof AreaFrame )
					{
						( (AreaFrame) bodyFrame ).setParent( singleContainer );
					}

					if ( DEBUG )
					{
						System.out.println( "------------aggregate sheet[" //$NON-NLS-1$
								+ ( currentPageIndex )
								+ "] using " //$NON-NLS-1$
								+ ( System.currentTimeMillis( ) - span )
								+ " ms" ); //$NON-NLS-1$
					}
				}
				else
				{
					// export body frame
					Sheet modelSheet = Transformer.toSheet( bodyFrame,
							1000,
							fixedColumnWidth * 1000 );
					exportSheet( modelSheet );

					if ( DEBUG )
					{
						System.out.println( "------------export sheet[" //$NON-NLS-1$
								+ ( currentPageIndex )
								+ "] using " //$NON-NLS-1$
								+ ( System.currentTimeMillis( ) - span )
								+ " ms" ); //$NON-NLS-1$
					}
				}
			}
			else
			{
				if ( exportSingleSheet )
				{
					// aggregate only
					if ( singleContainer == null )
					{
						singleContainer = new AggregateFrame( );
						singleContainer.setRight( currentFrame.getRight( )
								- currentFrame.getLeft( ) );
						singleContainer.setBottom( currentFrame.getBottom( )
								- currentFrame.getTop( ) );
					}
					else
					{
						int yOff = singleContainer.getBottom( );
						singleContainer.setBottom( yOff
								+ currentFrame.getBottom( )
								- currentFrame.getTop( ) );

						if ( currentFrame instanceof AreaFrame )
						{
							( (AreaFrame) currentFrame ).setYOffset( yOff );
						}
					}

					singleContainer.addChild( currentFrame );
					if ( currentFrame instanceof AreaFrame )
					{
						( (AreaFrame) currentFrame ).setParent( singleContainer );
					}

					if ( DEBUG )
					{
						System.out.println( "------------aggregate sheet[" //$NON-NLS-1$
								+ ( currentPageIndex )
								+ "] using " //$NON-NLS-1$
								+ ( System.currentTimeMillis( ) - span )
								+ " ms" ); //$NON-NLS-1$
					}
				}
				else
				{
					// one page completed, process it.
					Sheet modelSheet = Transformer.toSheet( currentFrame,
							1000,
							fixedColumnWidth * 1000 );
					exportSheet( modelSheet );

					if ( DEBUG )
					{
						System.out.println( "------------export sheet[" //$NON-NLS-1$
								+ ( currentPageIndex )
								+ "] using " //$NON-NLS-1$
								+ ( System.currentTimeMillis( ) - span )
								+ " ms" ); //$NON-NLS-1$
					}
				}
			}
		}
		currentPageIndex++;

	}

	/**
	 * Find child frame by specific data object(use "==" to compare)
	 * 
	 * @param parentFrame
	 * @param data
	 * @return
	 */
	protected Frame locateChildFrame( Frame parentFrame, Object data )
	{
		for ( Iterator<Frame> itr = parentFrame.iterator( ); itr.hasNext( ); )
		{
			Frame fr = itr.next( );

			if ( fr.getData( ) == data )
			{
				return fr;
			}

			Frame cfr = locateChildFrame( fr, data );
			if ( cfr != null )
			{
				return cfr;
			}
		}

		return null;
	}

	protected boolean isEffectiveCell( Cell cell )
	{
		// TODO currently birt style impl always return false for isEmpty(), so
		// the remove_emppry_row will not work, may need think about another
		// solution for this check
		if ( cell.isEmpty( ) )
		{
			return false;
		}

		return true;
	}

	protected void exportSheet( Sheet modelSheet )
	{
		HSSFSheet xlsSheet = workbook.createSheet( );
		HSSFPatriarch patriarch = xlsSheet.createDrawingPatriarch( );
		xlsSheet.setDisplayGridlines( showGridLines );

		// workbook.setSheetName( workbook.getNumberOfSheets( ) - 1,
		// modelSheet.getName( ),
		// HSSFWorkbook.ENCODING_UTF_16 );

		short columnCount = (short) ( Math.min( modelSheet.getActiveColumnRange( ),
				254 ) + 1 );
		int rowCount = modelSheet.getActiveRowRange( ) + 1;

		if ( columnCount + COMMENTS_WIDTH_IN_COLUMN < 255 )
		{
			commentsAnchorColumnIndex = columnCount;
		}
		else
		{
			commentsAnchorColumnIndex = 255 - COMMENTS_WIDTH_IN_COLUMN;
		}

		if ( rowCount + COMMENTS_HEIGHT_IN_ROW < 65535 )
		{
			commentsAnchorRowIndex = rowCount;
		}
		else
		{
			commentsAnchorRowIndex = 65535 - COMMENTS_HEIGHT_IN_ROW;
		}

		boolean[] nonBlankRow = new boolean[rowCount];

		if ( removeEmptyRow )
		{
			// check blank rows.
			for ( int i = 0; i < rowCount; i++ )
			{
				for ( int j = 0; j < columnCount; j++ )
				{
					Cell cell = modelSheet.getCell( i, j, false );

					if ( cell != null && isEffectiveCell( cell ) )
					{
						nonBlankRow[i] = true;
						break;
					}
				}
			}
		}

		double width = 0;
		for ( short i = 1; i < columnCount + 1; i++ )
		{
			width = modelSheet.getColumnWidth( i - 1 )
					/ ( 1000 * baseCharWidth );

			xlsSheet.setColumnWidth( (short) ( i - 1 ), (short) ( width * 256 ) );
		}

		HSSFCellStyle emptyCellStyle = processor.getEmptyCellStyle( );
		HSSFCell emptyCell = null;
		for ( short y = 0; y < rowCount; y++ )
		{
			if ( !removeEmptyRow || nonBlankRow[y] )
			{
				HSSFRow xlsRow = xlsSheet.createRow( y );
				double height = modelSheet.getRowHeight( y ) / 1000d;// + 2;

				// System.out.println( "row height " + y + ": " + height );

				xlsRow.setHeightInPoints( (float) height );

				for ( short x = 0; x < columnCount; x++ )
				{
					emptyCell = xlsRow.createCell( x );
					emptyCell.setCellStyle( emptyCellStyle );

					Cell element = modelSheet.getCell( y, x, false );
					if ( element != null )
					{

						exportCell( element,
								x,
								y,
								xlsSheet,
								patriarch,
								xlsRow,
								modelSheet );
					}
					else
					{
						emptyCell = xlsRow.createCell( x );
						emptyCell.setCellStyle( emptyCellStyle );
					}
				}
			}
			else
			{
				HSSFRow row = xlsSheet.createRow( y );
				row.setHeight( (short) 0 );
				for ( short x = 0; x < columnCount; x++ )
				{
					emptyCell = row.createCell( x );
					emptyCell.setCellStyle( emptyCellStyle );
				}
			}
		}
	}

	protected void exportCell( Cell element, short x, short y,
			HSSFSheet xlsSheet, HSSFPatriarch patriarch, HSSFRow xlsRow,
			Sheet modelSheet )
	{
		if ( element.isEmpty( ) )
		{
			return;
		}

		short x2 = x;
		short y2 = y;

		if ( element.isMerged( ) )
		{
			Iterator<MergeBlock> it = modelSheet.mergesIterator( );
			while ( it.hasNext( ) )
			{
				MergeBlock merge = it.next( );
				if ( merge.include( y, x ) )
				{
					if ( x == merge.getStartColumn( )
							&& y == merge.getStartRow( ) )
					{
						x2 = (short) ( x + merge.getColumnSpan( ) );
						y2 = (short) ( y + merge.getRowSpan( ) );
						xlsSheet.addMergedRegion( new Region( y, x, y2, x2 ) );

						break;
					}
				}
			}
		}

		HSSFCell cell = xlsRow.createCell( x );

		Object cellValue = getTranslatedElementValue( element.getValue( ) );
		boolean useHyperLinkStyle = false;

		if ( cellValue instanceof ITextArea )
		{
			useHyperLinkStyle = handleHyperLink( (IArea) cellValue, cell );
		}

		HSSFCellStyle cellStyle = processor.getHssfCellStyle( element.getStyle( ),
				useHyperLinkStyle );

		cell.setCellStyle( cellStyle );

		exportCellData( element, x, y, x2, y2, cell, patriarch );
	}

	protected Object getTranslatedElementValue( Object value )
	{
		return value;
	}

	protected void exportCellData( Cell element, short x, short y, short x2,
			short y2, HSSFCell cell, HSSFPatriarch patriarch )
	{
		Object cellValue = getTranslatedElementValue( element.getValue( ) );

		if ( cellValue instanceof IImageArea )
		{
			exportImage( (IImageArea) cellValue, cell, x, y, x2, y2, patriarch );
		}
		else if ( cellValue instanceof ITextArea )
		{
			exportText( (ITextArea) cellValue, cell );
		}

		if ( enableComments && cellValue instanceof IArea )
		{
			handleComments( (IArea) cellValue, cell, patriarch );
		}
	}

	protected void exportImage( IImageArea image, HSSFCell cell, short x,
			short y, short x2, short y2, HSSFPatriarch patriarch )
	{
		int idx = loadPicture( image );

		if ( idx != -1 )
		{
			HSSFClientAnchor anchor = new HSSFClientAnchor( 0,
					0,
					0,
					0,
					x,
					y,
					(short) ( x2 + 1 ),
					y2 + 1 );
			anchor.setAnchorType( 2 );
			HSSFPicture pic = patriarch.createPicture( anchor, idx );
			// pic.resize( );
		}
		else if ( !suppressUnknownImage )
		{
			cell.setCellValue( new HSSFRichTextString( "<<Unsupported Image>>" ) ); //$NON-NLS-1$
		}
	}

	protected void exportText( ITextArea text, HSSFCell cell )
	{
		Double csNumberValue = null;

		String csCellText = text.getText( );

		try
		{
			csNumberValue = Double.parseDouble( csCellText );
		}
		catch ( NumberFormatException e )
		{
			csNumberValue = null;
		}

		if ( csNumberValue == null || csNumberValue.isNaN( ) )
		{
			cell.setCellValue( new HSSFRichTextString( csCellText ) );
		}
		else
		{
			cell.setCellValue( csNumberValue );
		}
	}

	protected boolean handleHyperLink( IArea area, HSSFCell cell )
	{
		IContent content = area.getContent( );

		if ( content != null )
		{
			IHyperlinkAction hlAction = content.getHyperlinkAction( );

			if ( hlAction != null )
			{
				try
				{
					IReportRunnable runnable = services.getReportRunnable( );
					String systemId = runnable == null ? null
							: runnable.getReportName( );

					Action act = new Action( systemId, hlAction );

					String link = null;
					String tooltip = EngineUtil.getActionTooltip( hlAction );

					Object ac = services.getOption( RenderOption.ACTION_HANDLER );

					if ( ac instanceof IHTMLActionHandler )
					{
						link = ( (IHTMLActionHandler) ac ).getURL( act,
								services.getReportContext( ) );
					}
					else
					{
						link = hlAction.getHyperlink( );
					}

					if ( link != null )
					{
						HSSFHyperlink hssflink = new HSSFHyperlink( HSSFHyperlink.LINK_URL );
						hssflink.setAddress( link );
						if ( tooltip != null )
						{
							hssflink.setLabel( tooltip );
						}
						cell.setHyperlink( hssflink );

						return true;
					}
				}
				catch ( Exception e )
				{
					System.out.println( "create hyperlink failed." ); //$NON-NLS-1$

					e.printStackTrace( );
				}
			}
		}

		return false;
	}

	protected boolean handleComments( IArea area, HSSFCell cell,
			HSSFPatriarch patriarch )
	{
		IContent content = area.getContent( );

		if ( content != null )
		{
			Object sourceObj = content.getGenerateBy( );

			if ( sourceObj instanceof ReportElementDesign )
			{
				sourceObj = ( (ReportElementDesign) sourceObj ).getHandle( );

				if ( sourceObj instanceof ReportElementHandle )
				{
					String comments = ( (ReportElementHandle) sourceObj ).getComments( );

					if ( comments != null && comments.length( ) > 0 )
					{
						HSSFComment hssfcomment = patriarch.createComment( new HSSFClientAnchor( 0,
								0,
								0,
								0,
								(short) commentsAnchorColumnIndex,
								commentsAnchorRowIndex,
								(short) ( commentsAnchorColumnIndex + COMMENTS_WIDTH_IN_COLUMN ),
								commentsAnchorRowIndex + COMMENTS_HEIGHT_IN_ROW ) );
						hssfcomment.setString( new HSSFRichTextString( comments ) );

						cell.setCellComment( hssfcomment );

						return true;
					}
				}
			}
		}

		return false;
	}

	protected int loadPicture( IImageArea imageArea )
	{
		byte[] data = null;

		switch ( ( (IImageContent) imageArea.getContent( ) ).getImageSource( ) )
		{
			case IImageContent.IMAGE_FILE :
			case IImageContent.IMAGE_URL :
				String url = ( (IImageContent) imageArea.getContent( ) ).getURI( );
				if ( url != null )
				{
					try
					{
						BufferedInputStream bis = new BufferedInputStream( new URL( url ).openStream( ) );
						ByteArrayOutputStream bos = new ByteArrayOutputStream( );

						byte[] buf = new byte[1024];
						int count = bis.read( buf );
						while ( count != -1 )
						{
							bos.write( buf, 0, count );

							count = bis.read( buf );
						}

						data = bos.toByteArray( );
					}
					catch ( Exception e )
					{
						data = null;
					}
				}
				break;
			case IImageContent.IMAGE_NAME :
			case IImageContent.IMAGE_EXPRESSION :
				data = ( (IImageContent) imageArea.getContent( ) ).getData( );
				break;
		}

		if ( data != null )
		{
			int type = processor.getHssfPictureType( data );
			if ( type != -1 )
			{
				return workbook.addPicture( data, type );
			}
			else
			{
				try
				{
					// try convert to png format
					data = ImageUtil.convertImage( data, "png" ); //$NON-NLS-1$

					return workbook.addPicture( data,
							HSSFWorkbook.PICTURE_TYPE_PNG );
				}
				catch ( IOException e )
				{
				}
			}
		}

		return -1;
	}

	public void visitText( ITextArea textArea )
	{
		buildFrame( textArea, false );
	}

	public void visitImage( IImageArea imageArea )
	{
		buildFrame( imageArea, false );
	}

	public void visitAutoText( ITemplateArea templateArea )
	{
		buildFrame( templateArea, false );
	}

	public void visitContainer( IContainerArea containerArea )
	{
		startContainer( containerArea );

		Iterator iter = containerArea.getChildren( );
		while ( iter.hasNext( ) )
		{
			IArea child = (IArea) iter.next( );
			child.accept( this );
		}

		endContainer( containerArea );
	}

	public void setTotalPage( ITextArea totalPage )
	{
	}

	protected void buildFrame( IArea area, boolean isContainer )
	{
		AreaFrame frame = new AreaFrame( area );

		AreaFrame parentFrame = frameStack.isEmpty( ) ? null
				: (AreaFrame) frameStack.peek( );
		frame.setParent( parentFrame );
		if ( parentFrame != null )
		{
			parentFrame.addChild( frame );
		}

		if ( isContainer )
		{
			frameStack.push( frame );
		}

		// System.out.println(frame);
	}

}