/********************************************************************************
 * (C) Copyright 2000-2008, by Shawn Qualia.
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this library; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 ********************************************************************************/

package org.uguess.birt.report.engine.emitter.xls;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.eclipse.birt.report.engine.content.IContent;
import org.eclipse.birt.report.engine.content.IImageContent;
import org.eclipse.birt.report.engine.content.IReportContent;
import org.eclipse.birt.report.engine.content.IStyle;
import org.eclipse.birt.report.engine.content.ITextContent;
import org.eclipse.birt.report.engine.layout.area.IArea;
import org.eclipse.birt.report.engine.layout.area.IAreaVisitor;
import org.eclipse.birt.report.engine.layout.area.IContainerArea;
import org.eclipse.birt.report.engine.layout.area.IImageArea;
import org.eclipse.birt.report.engine.layout.area.ITextArea;
import org.eclipse.birt.report.engine.layout.area.impl.PageArea;
import org.eclipse.birt.report.engine.layout.pdf.font.FontInfo;
import org.uguess.birt.report.engine.layout.wrapper.Frame;
import org.uguess.birt.report.engine.layout.wrapper.impl.MultiAreaFrame;

/**
 * XlsRenderer2
 */
public class XlsRenderer2 extends XlsRenderer
{

	private Map<IContent, MultiAreaFrame> frameCache;
	private Stack<IArea> areaStack;
	private Stack<IArea> clipStack;
	private IArea bodyArea;

	@Override
	public void start( IReportContent rc )
	{
		super.start( rc );

		frameCache = new HashMap<IContent, MultiAreaFrame>( );
		areaStack = new Stack<IArea>( );
		clipStack = new Stack<IArea>( );
		bodyArea = null;
	}

	@Override
	public void end( IReportContent rc )
	{
		super.end( rc );

		frameCache.clear( );
		frameCache = null;

		areaStack.clear( );
		areaStack = null;

		clipStack.clear( );
		clipStack = null;

		bodyArea = null;
	}

	@Override
	public void startContainer( IContainerArea container )
	{
		if ( container.needClip( ) )
		{
			clipStack.push( wrap2global( container ) );
		}

		super.startContainer( container );

		areaStack.push( container );
	}

	@Override
	public void endContainer( IContainerArea containerArea )
	{
		if ( containerArea.needClip( ) )
		{
			clipStack.pop( );
		}

		areaStack.pop( );

		Frame currentFrame = frameStack.peek( );

		IArea area = unWrap( currentFrame.getData( ) );

		if ( area != containerArea )
		{
			return;
		}

		frameStack.pop( );

		if ( area instanceof PageArea )
		{
			resolveUnwantedContent( currentFrame );

			exportFrame( currentFrame );

			// clean up pagewise frame cache
			frameCache.clear( );
		}
	}

	private void resolveUnwantedContent( Frame frame )
	{
		Object cellValue = unWrapWithType( frame.getData( ) );

		int cnt = ( (MultiAreaFrame) frame ).getChildCount( );

		if ( cnt == 0 )
		{
			return;
		}

		if ( cellValue instanceof IImageArea )
		{
			// TODO could be action content, which is not supported now
			( (MultiAreaFrame) frame ).removeAllChildren( );
			return;
		}
		else if ( cellValue instanceof ITextArea )
		{
			// TODO could be action content, which is not supported now
			( (MultiAreaFrame) frame ).removeAllChildren( );
			return;
		}

		for ( Iterator<Frame> itr = frame.iterator( ); itr.hasNext( ); )
		{
			resolveUnwantedContent( itr.next( ) );
		}
	}

	@SuppressWarnings("unchecked")
	private IArea unWrap( Object data )
	{
		if ( data instanceof AreaWrapper )
		{
			return ( (AreaWrapper) data ).area;
		}

		if ( data instanceof IArea )
		{
			return (IArea) data;
		}

		if ( data instanceof List )
		{
			return unWrap( ( (List) data ).get( 0 ) );
		}

		return null;
	}

	@SuppressWarnings("unchecked")
	private IArea unWrapWithType( Object data )
	{
		if ( data instanceof AreaWrapper )
		{
			return ( (AreaWrapper) data ).area;
		}

		if ( data instanceof IArea )
		{
			return (IArea) data;
		}

		if ( data instanceof List )
		{
			return wrapWithType( (List) data );
		}

		return null;
	}

	private IArea wrapWithType( List<IArea> areas )
	{
		IArea hostArea = areas.get( 0 );

		if ( hostArea.getContent( ) instanceof ITextContent )
		{
			for ( int i = 0; i < areas.size( ); i++ )
			{
				IArea ar = unWrap( areas.get( i ) );

				if ( ar instanceof ITextArea )
				{
					TextAreaWrapper textWrapper = new TextAreaWrapper( (ITextArea) ar );
					textWrapper.text = ( (ITextContent) hostArea.getContent( ) ).getText( );

					return textWrapper;
				}
			}
		}
		else if ( hostArea.getContent( ) instanceof IImageContent )
		{
			for ( int i = 0; i < areas.size( ); i++ )
			{
				IArea ar = unWrap( areas.get( i ) );

				if ( ar instanceof IImageArea )
				{
					// return first image
					return ar;
				}
			}
		}

		return hostArea;
	}

	@Override
	protected Object getTranslatedElementValue( Object value )
	{
		return unWrapWithType( value );
	}

	@Override
	protected void buildFrame( IArea area, boolean isContainer )
	{
		IContent cnt = area.getContent( );

		if ( area instanceof PageArea )
		{
			// record body area to support export_body_only option
			bodyArea = ( (PageArea) area ).getBody( );
		}

		if ( cnt != null )
		{
			MultiAreaFrame frame = frameCache.get( cnt );

			if ( frame != null )
			{
				frame.addSubArea( wrap2global( area ) );
			}
			else
			{
				frame = buildNewFrame( wrap2global( area ), isContainer );

				frameCache.put( cnt, frame );
			}
		}
		else
		{
			if ( area == bodyArea )
			{
				// create body frame
				buildNewFrame( wrap2global( area ), isContainer );
				bodyArea = null;
			}
		}
	}

	@Override
	protected Frame locateChildFrame( Frame parentFrame, Object data )
	{
		for ( Iterator<Frame> itr = parentFrame.iterator( ); itr.hasNext( ); )
		{
			Frame fr = itr.next( );

			if ( unWrap( fr.getData( ) ) == data )
			{
				return fr;
			}

			Frame cfr = locateChildFrame( fr, data );
			if ( cfr != null )
			{
				return cfr;
			}
		}

		return null;
	}

	private IArea wrap2global( IArea child )
	{
		if ( areaStack.isEmpty( ) )
		{
			return child;
		}

		AreaWrapper warea = new AreaWrapper( child );

		for ( int i = areaStack.size( ) - 1; i >= 0; i-- )
		{
			IArea stackArea = areaStack.get( i );

			warea.x += stackArea.getX( );
			warea.y += stackArea.getY( );
		}

		return warea;
	}

	private MultiAreaFrame buildNewFrame( IArea area, boolean isContainer )
	{
		MultiAreaFrame frame = new MultiAreaFrame( area );

		if ( !clipStack.isEmpty( ) )
		{
			frame.setClip( clipStack.peek( ) );
		}

		MultiAreaFrame parentFrame = frameStack.isEmpty( ) ? null
				: (MultiAreaFrame) frameStack.peek( );
		frame.setParent( parentFrame );

		if ( parentFrame != null )
		{
			parentFrame.addChild( frame );
		}

		if ( isContainer )
		{
			frameStack.push( frame );
		}

		// System.out.println(frame);

		return frame;
	}

	/**
	 * AreaWrapper
	 */
	static class AreaWrapper implements IArea
	{

		protected IArea area;
		private int x, y;

		AreaWrapper( IArea area )
		{
			this.area = area;

			x = area.getX( );
			y = area.getY( );
		}

		public void accept( IAreaVisitor visitor )
		{
			area.accept( visitor );
		}

		public IContent getContent( )
		{
			return area.getContent( );
		}

		public int getHeight( )
		{
			return area.getHeight( );
		}

		public float getScale( )
		{
			return area.getScale( );
		}

		public IStyle getStyle( )
		{
			return area.getStyle( );
		}

		public int getWidth( )
		{
			return area.getWidth( );
		}

		public int getX( )
		{
			return x;
		}

		public int getY( )
		{
			return y;
		}
	}

	/**
	 * TextAreaWrapper
	 */
	static class TextAreaWrapper extends AreaWrapper implements ITextArea
	{

		private String text;
		private String logicText;

		private static Method mtdGetLogicalOrderText;

		static
		{
			try
			{
				mtdGetLogicalOrderText = ITextArea.class.getDeclaredMethod( "getLogicalOrderText", //$NON-NLS-1$
						(Class[]) null );
			}
			catch ( Exception e )
			{
				mtdGetLogicalOrderText = null;

				e.printStackTrace( );
			}
		}

		TextAreaWrapper( ITextArea area )
		{
			super( area );

			this.text = area.getText( );

			if ( mtdGetLogicalOrderText != null )
			{
				try
				{
					this.logicText = (String) mtdGetLogicalOrderText.invoke( area,
							(Object[]) null );
				}
				catch ( Exception e )
				{
					e.printStackTrace( );
				}
			}
			else
			{
				this.logicText = this.text;
			}
		}

		public FontInfo getFontInfo( )
		{
			return ( (ITextArea) area ).getFontInfo( );
		}

		public String getText( )
		{
			return text;
		}

		public String getLogicalOrderText( )
		{
			return logicText;
		}
	}
}
