/********************************************************************************
 * (C) Copyright 2000-2008, by Shawn Qualia.
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this library; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 ********************************************************************************/

package org.uguess.birt.report.engine.emitter.xls;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFDataFormat;
import org.apache.poi.hssf.usermodel.HSSFFont;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.eclipse.birt.report.engine.content.IStyle;
import org.eclipse.birt.report.engine.css.engine.StyleConstants;
import org.eclipse.birt.report.engine.css.engine.value.css.CSSConstants;
import org.eclipse.birt.report.engine.layout.pdf.util.PropertyUtil;
import org.uguess.birt.report.engine.util.ImageUtil;

/**
 * XlsStyleProcessor
 */
public class XlsStyleProcessor
{

	private HSSFWorkbook workbook;

	private HSSFCellStyle emptyCellStyle;

	private Map<?, HSSFColor> hssfColorMap;

	private List<HSSFFont> fontCache;

	private Map<IStyle, HSSFCellStyle[]> styleCache;

	@SuppressWarnings("unchecked")
	XlsStyleProcessor( HSSFWorkbook workbook )
	{
		this.workbook = workbook;

		this.fontCache = new ArrayList<HSSFFont>( );
		this.styleCache = new HashMap<IStyle, HSSFCellStyle[]>( );
		this.hssfColorMap = HSSFColor.getIndexHash( );

		initEmptyCellStyle( );
	}

	public void dispose( )
	{
		fontCache.clear( );
		styleCache.clear( );
		hssfColorMap.clear( );

		fontCache = null;
		styleCache = null;
		hssfColorMap = null;
		emptyCellStyle = null;

		workbook = null;
	}

	private void initEmptyCellStyle( )
	{
		emptyCellStyle = workbook.createCellStyle( );

		emptyCellStyle.setFillPattern( HSSFCellStyle.SOLID_FOREGROUND );
		emptyCellStyle.setAlignment( HSSFCellStyle.ALIGN_CENTER );
		emptyCellStyle.setHidden( false );
		emptyCellStyle.setLocked( false );
		emptyCellStyle.setBorderLeft( HSSFCellStyle.BORDER_NONE );
		emptyCellStyle.setBorderRight( HSSFCellStyle.BORDER_NONE );
		emptyCellStyle.setBorderTop( HSSFCellStyle.BORDER_NONE );
		emptyCellStyle.setBorderBottom( HSSFCellStyle.BORDER_NONE );

		short backColorIndex = HSSFColor.BLACK.index;
		emptyCellStyle.setBottomBorderColor( backColorIndex );
		emptyCellStyle.setTopBorderColor( backColorIndex );
		emptyCellStyle.setLeftBorderColor( backColorIndex );
		emptyCellStyle.setRightBorderColor( backColorIndex );
		emptyCellStyle.setFillForegroundColor( HSSFColor.WHITE.index );
		emptyCellStyle.setFillBackgroundColor( HSSFColor.WHITE.index );
		emptyCellStyle.setVerticalAlignment( HSSFCellStyle.VERTICAL_CENTER );

		HSSFFont defaultFont = workbook.createFont( );
		defaultFont.setFontName( "Serif" ); //$NON-NLS-1$
		short size = 10;
		defaultFont.setFontHeightInPoints( size );
		emptyCellStyle.setFont( defaultFont );
	}

	public HSSFCellStyle getEmptyCellStyle( )
	{
		return emptyCellStyle;
	}

	public int getHssfPictureType( byte[] data )
	{
		int type = ImageUtil.getImageType( data );

		switch ( type )
		{
			case ImageUtil.TYPE_DIB :
				return HSSFWorkbook.PICTURE_TYPE_DIB;
			case ImageUtil.TYPE_PNG :
				return HSSFWorkbook.PICTURE_TYPE_PNG;
			case ImageUtil.TYPE_JPEG :
				return HSSFWorkbook.PICTURE_TYPE_JPEG;
		}

		return -1;
	}

	public HSSFCellStyle getHssfCellStyle( IStyle style, boolean useLinkStyle )
	{
		HSSFCellStyle hssfStyle = null;

		if ( style.isEmpty( ) )
		{
			return emptyCellStyle;
		}

		HSSFCellStyle[] styleEntry = null;

		// check existing cell style cache first.
		for ( Iterator<Entry<IStyle, HSSFCellStyle[]>> itr = styleCache.entrySet( )
				.iterator( ); itr.hasNext( ); )
		{
			Entry<IStyle, HSSFCellStyle[]> entry = itr.next( );

			if ( style.equals( entry.getKey( ) ) )
			{
				styleEntry = entry.getValue( );
				hssfStyle = styleEntry[useLinkStyle ? 1 : 0];
				break;
			}
		}

		if ( hssfStyle == null )
		{
			hssfStyle = workbook.createCellStyle( );

			hssfStyle.setFont( getHssfFont( style, useLinkStyle ) );

			Color color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_BACKGROUND_COLOR ) );
			if ( color != null )
			{
				short cdx = getHssfColor( color ).getIndex( );
				hssfStyle.setFillPattern( HSSFCellStyle.SOLID_FOREGROUND );
				hssfStyle.setFillBackgroundColor( cdx );
				hssfStyle.setFillForegroundColor( cdx );
			}

			if ( style.getBorderLeftStyle( ) != null
					&& !"none".equals( style.getBorderLeftStyle( ) ) ) //$NON-NLS-1$
			{
				hssfStyle.setBorderLeft( getHssfBorder( PropertyUtil.getDimensionValue( style.getProperty( StyleConstants.STYLE_BORDER_LEFT_WIDTH ) ),
						style.getBorderLeftStyle( ) ) );

				if ( hssfStyle.getBorderLeft( ) != HSSFCellStyle.BORDER_NONE )
				{
					color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_BORDER_LEFT_COLOR ) );
					if ( color != null )
					{
						hssfStyle.setLeftBorderColor( getHssfColor( color ).getIndex( ) );
					}
				}
			}
			if ( style.getBorderRightStyle( ) != null
					&& !"none".equals( style.getBorderRightStyle( ) ) ) //$NON-NLS-1$
			{
				hssfStyle.setBorderRight( getHssfBorder( PropertyUtil.getDimensionValue( style.getProperty( StyleConstants.STYLE_BORDER_RIGHT_WIDTH ) ),
						style.getBorderRightStyle( ) ) );

				if ( hssfStyle.getBorderRight( ) != HSSFCellStyle.BORDER_NONE )
				{
					color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_BORDER_RIGHT_COLOR ) );
					if ( color != null )
					{
						hssfStyle.setRightBorderColor( getHssfColor( color ).getIndex( ) );
					}
				}
			}
			if ( style.getBorderTopStyle( ) != null
					&& !"none".equals( style.getBorderTopStyle( ) ) ) //$NON-NLS-1$
			{
				hssfStyle.setBorderTop( getHssfBorder( PropertyUtil.getDimensionValue( style.getProperty( StyleConstants.STYLE_BORDER_TOP_WIDTH ) ),
						style.getBorderTopStyle( ) ) );

				if ( hssfStyle.getBorderTop( ) != HSSFCellStyle.BORDER_NONE )
				{
					color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_BORDER_TOP_COLOR ) );
					if ( color != null )
					{
						hssfStyle.setTopBorderColor( getHssfColor( color ).getIndex( ) );
					}
				}
			}
			if ( style.getBorderBottomStyle( ) != null
					&& !"none".equals( style.getBorderBottomStyle( ) ) ) //$NON-NLS-1$
			{
				hssfStyle.setBorderBottom( getHssfBorder( PropertyUtil.getDimensionValue( style.getProperty( StyleConstants.STYLE_BORDER_BOTTOM_WIDTH ) ),
						style.getBorderBottomStyle( ) ) );

				if ( hssfStyle.getBorderBottom( ) != HSSFCellStyle.BORDER_NONE )
				{
					color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_BORDER_BOTTOM_COLOR ) );
					if ( color != null )
					{
						hssfStyle.setBottomBorderColor( getHssfColor( color ).getIndex( ) );
					}
				}
			}

			hssfStyle.setAlignment( getHssfAlign( style.getTextAlign( ), true ) );
			hssfStyle.setVerticalAlignment( getHssfAlign( style.getVerticalAlign( ),
					false ) );
			hssfStyle.setWrapText( true );

			if ( style.getNumberFormat( ) != null
					&& style.getNumberFormat( ).length( ) > 0 )
			{
				short builtInFormat = HSSFDataFormat.getBuiltinFormat( style.getNumberFormat( ) );
				if ( builtInFormat != -1 )
				{
					hssfStyle.setDataFormat( builtInFormat );
				}
			}

			if ( styleEntry == null )
			{
				styleEntry = new HSSFCellStyle[2];
				styleEntry[useLinkStyle ? 1 : 0] = hssfStyle;

				styleCache.put( style, styleEntry );
			}
			else
			{
				styleEntry[useLinkStyle ? 1 : 0] = hssfStyle;
			}
		}
		return hssfStyle;
	}

	private HSSFFont getHssfFont( IStyle style, boolean useLinkStyle )
	{
		HSSFFont cellFont = null;

		short forecolor;
		if ( useLinkStyle )
		{
			forecolor = HSSFColor.BLUE.index;
		}
		else
		{
			Color color = PropertyUtil.getColor( style.getProperty( StyleConstants.STYLE_COLOR ) );
			forecolor = color == null ? HSSFColor.BLACK.index
					: getHssfColor( color ).getIndex( );
		}

		String fontName = style.getFontFamily( ) == null ? "Serif" //$NON-NLS-1$
				: style.getFontFamily( );
		short fontSize = (short) ( PropertyUtil.getDimensionValue( style.getProperty( StyleConstants.STYLE_FONT_SIZE ) ) / 1000d );
		byte underline = useLinkStyle ? ( HSSFFont.U_SINGLE )
				: ( "underline".equalsIgnoreCase( style.getTextUnderline( ) ) ? HSSFFont.U_SINGLE //$NON-NLS-1$
						: HSSFFont.U_NONE );
		boolean strikeout = "line-through".equalsIgnoreCase( style.getTextLineThrough( ) ); //$NON-NLS-1$
		short boldweight = PropertyUtil.isBoldFont( style.getProperty( StyleConstants.STYLE_FONT_WEIGHT ) ) ? HSSFFont.BOLDWEIGHT_BOLD
				: HSSFFont.BOLDWEIGHT_NORMAL;
		boolean italic = ( CSSConstants.CSS_OBLIQUE_VALUE.equals( style.getFontStyle( ) ) || CSSConstants.CSS_ITALIC_VALUE.equals( style.getFontStyle( ) ) );

		// search cache to reuse existing same font.
		for ( Iterator<HSSFFont> itr = fontCache.iterator( ); itr.hasNext( ); )
		{
			HSSFFont font = itr.next( );

			if ( font.getColor( ) == forecolor
					&& font.getFontName( ).equals( fontName )
					&& font.getFontHeightInPoints( ) == fontSize
					&& font.getUnderline( ) == underline
					&& font.getStrikeout( ) == strikeout
					&& font.getBoldweight( ) == boldweight
					&& font.getItalic( ) == italic )
			{
				cellFont = font;
				break;
			}
		}

		if ( cellFont == null )
		{
			cellFont = workbook.createFont( );
			cellFont.setFontName( fontName );
			cellFont.setColor( forecolor );
			cellFont.setFontHeightInPoints( fontSize );
			cellFont.setUnderline( underline );
			cellFont.setStrikeout( strikeout );
			cellFont.setBoldweight( boldweight );
			cellFont.setItalic( italic );

			fontCache.add( cellFont );
		}
		return cellFont;
	}

	private HSSFColor getHssfColor( Color awtColor )
	{
		HSSFColor color = null;
		if ( hssfColorMap != null && hssfColorMap.size( ) > 0 )
		{
			HSSFColor crtColor = null;
			short[] rgb = null;
			int diff = 0;
			int minDiff = 999;

			for ( Iterator<HSSFColor> it = hssfColorMap.values( ).iterator( ); it.hasNext( ); )
			{
				crtColor = it.next( );
				rgb = crtColor.getTriplet( );
				diff = Math.abs( rgb[0] - awtColor.getRed( ) )
						+ Math.abs( rgb[1] - awtColor.getGreen( ) )
						+ Math.abs( rgb[2] - awtColor.getBlue( ) );

				if ( diff == 0 )
				{
					// precise match
					color = crtColor;
					break;
				}

				if ( diff < minDiff )
				{
					minDiff = diff;
					color = crtColor;
				}
			}
		}
		return color;
	}

	private short getHssfBorder( int borderWidth, String borderStyle )
	{
		int width = (int) ( ( borderWidth + 500 ) / 1000d );

		switch ( width )
		{
			case 0 :
				return HSSFCellStyle.BORDER_NONE;
			case 1 :
			{
				if ( "none".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_NONE;
				}
				if ( "double".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DOUBLE;
				}
				if ( "solid".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_THIN;
				}
				if ( "dashed".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DASHED;
				}
				if ( "dotted".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DOTTED;
				}
				return HSSFCellStyle.BORDER_THIN;
			}
			case 2 :
			default :
			{
				// equal or greater than 2.
				if ( "none".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_NONE;
				}
				if ( "double".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DOUBLE;
				}
				if ( "solid".equals( borderStyle ) ) //$NON-NLS-1$
				{
					if ( width == 2 )
					{
						return HSSFCellStyle.BORDER_MEDIUM;
					}
					else
					{
						return HSSFCellStyle.BORDER_THICK;
					}
				}
				if ( "dashed".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DASHED;
				}
				if ( "dotted".equals( borderStyle ) ) //$NON-NLS-1$
				{
					return HSSFCellStyle.BORDER_DOTTED;
				}
				if ( width == 2 )
				{
					return HSSFCellStyle.BORDER_MEDIUM;
				}
				else
				{
					return HSSFCellStyle.BORDER_THICK;
				}
			}
		}
	}

	private short getHssfAlign( String align, boolean horizontal )
	{
		if ( horizontal )
		{
			short horizontalAlignment = HSSFCellStyle.ALIGN_LEFT;

			if ( "center".equals( align ) ) //$NON-NLS-1$
			{
				horizontalAlignment = HSSFCellStyle.ALIGN_CENTER;
			}
			else if ( "right".equals( align ) ) //$NON-NLS-1$
			{
				horizontalAlignment = HSSFCellStyle.ALIGN_RIGHT;
			}
			else if ( "justify".equals( align ) ) //$NON-NLS-1$
			{
				horizontalAlignment = HSSFCellStyle.ALIGN_JUSTIFY;
			}
			return horizontalAlignment;
		}
		else
		{
			short verticalAlignment = HSSFCellStyle.VERTICAL_CENTER;

			if ( "top".equals( align ) ) //$NON-NLS-1$
			{
				verticalAlignment = HSSFCellStyle.VERTICAL_TOP;
			}
			else if ( "bottom".equals( align ) ) //$NON-NLS-1$
			{
				verticalAlignment = HSSFCellStyle.VERTICAL_BOTTOM;
			}
			else if ( "justify".equals( align ) ) //$NON-NLS-1$
			{
				verticalAlignment = HSSFCellStyle.VERTICAL_JUSTIFY;
			}
			return verticalAlignment;
		}
	}
}
